;;; smv-mode.el --- SMV code editing commands for Emacs

;; Copyright (C) 1988, 1994 Free Software Foundation, Inc.

;; Maintainer: FSF
;; Keywords: unix, languages

;; This file is part of XEmacs.

;; XEmacs is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; XEmacs is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with XEmacs; see the file COPYING.  If not, write to the Free
;; Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;; 02111-1307, USA.

;;; Synched up with: FSF 19.34.

;;; Commentary:

;; Sets up C-mode with support for smv-style #-comments and a lightly
;; hacked syntax table. This is derived from awk-mode.el.

;;; Code:

(require 'cc-mode)

(defvar smv-mode-syntax-table nil
  "Syntax table in use in Smv-mode buffers.")

(if smv-mode-syntax-table
    ()
  (setq smv-mode-syntax-table (make-syntax-table))
  ;; Populate the syntax TABLE
  ;; DO NOT TRY TO SET _ (UNDERSCORE) TO WORD CLASS!
  (modify-syntax-entry ?_  "_"     smv-mode-syntax-table)
  (modify-syntax-entry ?\\ "\\"    smv-mode-syntax-table)
  (modify-syntax-entry ?+  "."     smv-mode-syntax-table)
  (modify-syntax-entry ?-  "."     smv-mode-syntax-table)
  (modify-syntax-entry ?=  "."     smv-mode-syntax-table)
  (modify-syntax-entry ?%  "."     smv-mode-syntax-table)
  (modify-syntax-entry ?<  "."     smv-mode-syntax-table)
  (modify-syntax-entry ?>  "."     smv-mode-syntax-table)
  (modify-syntax-entry ?&  "."     smv-mode-syntax-table)
  (modify-syntax-entry ?|  "."     smv-mode-syntax-table)
  (modify-syntax-entry ?\' "\""    smv-mode-syntax-table)
  (modify-syntax-entry ?/  ". 14"  smv-mode-syntax-table)
  (modify-syntax-entry ?*  ". 23"  smv-mode-syntax-table)
)

(defvar smv-mode-abbrev-table nil
  "Abbrev table in use in Smv-mode buffers.")
(define-abbrev-table 'smv-mode-abbrev-table ())

(defvar smv-mode-map ()
  "Keymap used in smv-mode buffers.")
(if smv-mode-map
    nil
  (setq smv-mode-map (c-make-inherited-keymap))
  ;; add bindings which are only useful for C
  (define-key smv-mode-map "\C-c\C-r"  'vw-on-current-buffer)
  )

(defun vw-on-current-buffer ()
  (interactive)
  (and
   (buffer-modified-p)
   (y-or-n-p (format "Save file %s? " (buffer-file-name)))
   (save-buffer))
  (start-process "vw" nil "vw" (buffer-file-name)))

(defvar smv-font-lock-keywords nil "Default expressions to highlight in SMV mode")

;;;###autoload
(defun smv-mode ()
  "Major mode for editing SMV code, based on C mode.

The hook variable `smv-mode-hook' is run with no args, if that value is
bound and has a non-nil value.  Also the hook `smv-mode-common-hook' is
run first.

Key bindings:
\\{smv-mode-map}"
  (interactive)
  (require 'cc-mode)
  (c-initialize-cc-mode)
  (kill-all-local-variables)
  (set-syntax-table smv-mode-syntax-table)
  (setq major-mode 'smv-mode
	mode-name "SMV"
	local-abbrev-table c-mode-abbrev-table)
  (use-local-map smv-mode-map)
  (c-common-init)
  (add-menu-button '("SMV") '["Run VW on File"         vw-on-current-buffer t])
  (setq comment-start "/* "
	comment-end   " */"
	c-conditional-key c-C-conditional-key
	c-class-key c-C-class-key
	c-baseclass-key nil
	c-comment-start-regexp "/\*"
	c-bitfield-key c-C-bitfield-key
	)
  (cc-imenu-init cc-imenu-c-generic-expression)
  (run-hooks 'smv-mode-common-hook)
  (run-hooks 'smv-mode-hook)
  (c-update-modeline))



(let ((c-keywords
;      ("break" "continue" "do" "else" "for" "if" "return" "switch" "while")
       "assert\\|c\\(ase\\|hain\\)\\|do\\|else\\|for\\(\\|all\\)\\|if\\|prove\\|return\\|s\\(ubcase\\|witch\\)\\|using\\|while")
      (c-type-types
       "array\\|boolean\\|enum\\|module\\|o\\(f\\|rdset\\)\\|s\\(calarset\\|truct\\)\\|typedef")
;       (concat "auto\\|c\\(har\\|onst\\)\\|double\\|e\\(num\\|xtern\\)\\|"
;	       "float\\|int\\|long\\|register\\|"
;	       "s\\(hort\\|igned\\|t\\(atic\\|ruct\\)\\)\\|typedef\\|"
;	       "un\\(ion\\|signed\\)\\|vo\\(id\\|latile\\)"))	; 6 ()s deep.
      (ctoken "\\(\\sw\\|\\s_\\|[:~*&]\\)+")
      )
 (setq smv-font-lock-keywords
  (list
   ;;
   ;; These are all anchored at the beginning of line for speed.
   ;;
   ;; Fontify function name definitions (GNU style; without type on line).
   
   ;; In FSF this has the simpler definition of "\\sw+" for ctoken.
   ;; I'm not sure if ours is more correct.
   ;; This is a subset of the next rule, and is slower when present. --dmoore
   ;; (list (concat "^\\(" ctoken "\\)[ \t]*(") 1 'font-lock-function-name-face)
   ;;
   ;; fontify the names of functions being defined.
   ;; FSF doesn't have this but I think it should be fast for us because
   ;; our regexp routines are more intelligent than FSF's about handling
   ;; anchored-at-newline. (When I added this hack in regex.c, it halved
   ;; the time to do the regexp phase of font-lock for a C file!) Not
   ;; including this discriminates against those who don't follow the
   ;; GNU coding style. --ben
   ;; x?x?x?y?z should always be: (x(xx?)?)?y?z --dmoore
   (list (concat
	  "^\\("
          "\\(" ctoken "[ \t]+\\)"	; type specs; there can be no
	  "\\("
          "\\(" ctoken "[ \t]+\\)"	; more than 3 tokens, right?
          "\\(" ctoken "[ \t]+\\)"
	  "?\\)?\\)?"
          "\\([*&]+[ \t]*\\)?"		; pointer
          "\\(" ctoken "\\)[ \t]*(")	; name
         10 'font-lock-function-name-face)
   ;;
   ;; This is faster but not by much.  I don't see why not.
   ;(list (concat "^\\(" ctoken "\\)[ \t]*(") 1 'font-lock-function-name-face)
   ;;
   ;; Added next two; they're both jolly-good fastmatch candidates so
   ;; should be fast. --ben
   ;;
   ;; Fontify structure names (in structure definition form).
   (list (concat "^\\(typedef\\|scalarset\\|ordset\\)"
      	   "[ \t]+\\(" ctoken "\\)")
         2 'font-lock-function-name-face)
   ;;
   ;; Fontify filenames in #include <...> preprocessor directives as strings.
   '("^#[ \t]*include[ \t]+\\(<[^>\"\n]+>\\)" 1 font-lock-string-face)
   ;;
   ;; Fontify function macro names.
   '("^#[ \t]*define[ \t]+\\(\\(\\sw+\\)(\\)" 2 font-lock-function-name-face)
   ;;
   ;; Fontify symbol names in #if ... defined preprocessor directives.
   '("^#[ \t]*if\\>"
     ("\\<\\(defined\\)\\>[ \t]*(?\\(\\sw+\\)?" nil nil
      (1 font-lock-preprocessor-face) (2 font-lock-variable-name-face nil t)))
   ;;
   ;; Fontify symbol names in #elif ... defined preprocessor directives.
   '("^#[ \t]*elif\\>"
     ("\\<\\(defined\\)\\>[ \t]*(?\\(\\sw+\\)?" nil nil
      (1 font-lock-preprocessor-face) (2 font-lock-variable-name-face nil t)))
   ;;
   ;; Fontify otherwise as symbol names, and the preprocessor directive names.
   '("^\\(#[ \t]*[a-z]+\\)\\>[ \t]*\\(\\sw+\\)?"
     (1 font-lock-preprocessor-face) (2 font-lock-variable-name-face nil t))
   ;; Fontify all type specifiers.
   (cons (concat "\\<\\(" c-type-types "\\)\\>") 'font-lock-type-face)
   ;;
   ;; Fontify all builtin keywords (except case, default and goto; see below).
   (cons (concat "\\<\\(" c-keywords "\\)\\>") 'font-lock-keyword-face)
   )))


;;; smv-mode.el ends here
